import kappy
from .KappaClasses import KappaAgent, KappaSiteState
from .FormattedKappaError import FormattedKappaError

import time


class KappaModel:
    """Class to represent Kappa models and create corresponding Kappa scripts.

    See the README example for a quick intro.

    Parameters
    ----------

    agents
      List of all KappaAgent in the model.

    rules
      List of all KappaRule in the model.

    initial_quantities
      Either a dict {agent_name: initial_quantity} or a single integer
      if all agents start with the same initial quantity. Note that the
      more initial agents, the less noisy the simulation results/

    duration
      Virtual duration of the simulation, in seconds/

    snapshot_times
        A dict {snapshot_name: time}, e.g. {'end': 10}. The snapshot can then
        be accessed as simulation_results['snapshots']['end'].

    plots
      List of Kappa expressions to plot. Each list element can be either
      a string (a Kappa expression like ``|A(.)|``) or a KappaSiteState like
      ``[KappaSiteState('B', 'b', '.'), KappaSiteState('C', 'c', '.')]``.

    plot_time_step
      Time interval between two points when gathering data for plotting.

    stop_condition
      String representing a Kappa stopping condition. If none is provided,
      the simulation stops after the provided ``duration`` is reached.
    """

    def __init__(
        self,
        agents,
        rules,
        initial_quantities,
        snapshot_times=None,
        plots=(),
        plot_time_step=0.1,
        duration=None,
        stop_condition=None,
    ):

        self.agents = agents
        self.rules = rules
        self.initial_quantities = initial_quantities
        self.snapshot_times = snapshot_times or {}
        self.plots = list(plots)
        self.set_parameters(
            duration=duration,
            stop_condition=stop_condition,
            plot_time_step=plot_time_step,
        )

    def set_parameters(self, duration=None, stop_condition=None, plot_time_step=0.1):
        """(Re-)set some parameters of the model.

        Do not attempt to set these parameters otherwise than with this
        function, e.g. directly by accessing self.parameters: it won't work.
        """
        if stop_condition is None:
            stop_condition = "[T] > %.04f" % duration
        self.parameters = kappy.SimulationParameter(
            plot_period=plot_time_step, pause_condition=stop_condition
        )

    def _kappa_script_for_agents_declarations(self):
        """Generate the script for declaring agents."""
        return "\n".join([a._kappa_declaration() for a in self.agents])

    def _kappa_script_for_rules(self):
        """Generate the script for declaring all complexation rules."""
        return "\n".join([r._kappa() for r in self.rules])

    def _kappa_script_for_initial_quantities(self):
        """Generate the script for declaring all initial quantities."""
        return "\n".join(
            [
                "%%init: %d %s()"
                % (n, agent.name if isinstance(agent, KappaAgent) else agent)
                for agent, n in self.initial_quantities.items()
            ]
        )

    def _kappa_script_for_snapshots(self):
        """Generate the script for declaring when snapshots are recorded."""
        return "\n".join(
            [
                '%%mod: alarm %.03f do $SNAPSHOT "%s";' % (t, sid)
                for sid, t in self.snapshot_times.items()
            ]
        )

    def _auto_plot_item_string(self, item):
        """Generate the Kappa string for the number of agents or sites.

        This is mainly a helper for _kappa_script_for_plotted.
        """
        if isinstance(item, KappaAgent):
            return "|%s()|" % item.name
        elif isinstance(item, KappaSiteState):
            return "|%s|" % item._kappa()
        else:
            return item

    def _kappa_script_for_plotted(self):
        """Generate the script for declaring what gets plotted."""
        return "\n".join(
            [
                "%%plot: %s" % self._auto_plot_item_string(plot_item)
                for plot_item in self.plots
            ]
        )

    def _full_kappa_script(self):
        """Generate the full script to be passed to the Kappa simulator."""
        return "\n\n".join(
            [
                self._kappa_script_for_agents_declarations(),
                self._kappa_script_for_rules(),
                self._kappa_script_for_initial_quantities(),
                self._kappa_script_for_snapshots(),
                self._kappa_script_for_plotted(),
            ]
        )

    def get_simulation_results(self):
        """Run a simulation of the model and return results as a dict.

        The result is of the form {plots: {}, snapshots {}}.

        The plots dict is of the form {'[T]': [...], 'A()': [...]} where the
        values are lists of numbers.

        The snapshots dict is of the form {snapshot_name: data} where the
        data is a complicated listing of the different complexes observed at
        that time point, and how often they occur. (See kappy documentation
        for more).

        Topkappy has a methods like ``plot_simulation_time_series`` or
        ``plot_snapshot_agents`` to help make sense of the simulation
        results.
        """
        model_string = self._full_kappa_script()
        kappa_client = kappy.KappaStd()
        kappa_client.add_model_string(model_string)
        try:
            kappa_client.project_parse()
        except kappy.KappaError as kappa_error:
            raise FormattedKappaError.from_kappa_error(kappa_error, model_string)
        kappa_client.simulation_start(self.parameters)
        kappa_client.wait_for_simulation_stop()
        plot_data = kappa_client.simulation_plot()
        plot_data = dict(zip(plot_data["legend"], zip(*plot_data["series"])))
        snapshots = {}

        _ = kappa_client.simulation_snapshots()
        # print(kappa_client.__dict__)

        def get_shapshots():
            snapshots = {}
            for sid in list(self.snapshot_times) + ["deadlock"]:
                try:
                    name = sid + ".ka"
                    snapshots[sid] = kappa_client.simulation_snapshot(name)
                except:
                    try:
                        snapshots[sid] = kappa_client.simulation_snapshot(sid)
                    except:
                        pass
            return snapshots

        for _try in range(3):
            snapshots = get_shapshots()
            if len(snapshots):
                break
            time.sleep(0.2)

        return {"plots": plot_data, "snapshots": snapshots}



class KappaAgent:
    """Class to represent a Kappa agent.

    Parameters
    ----------
    name
      Agent name, e.g. 'A'.

    sites
      List of sites, e.g. ['a1', 'a2'].
    """

    def __init__(self, name, sites):
        self.name = name
        self.sites = sites

    def _kappa_declaration(self):
        return "%%agent: %s(%s)" % (self.name, ", ".join(self.sites))

    def __str__(self):
        return self._kappa_declaration()


class KappaSiteState:
    """Class to represent a Kappa site state like 'A(x[.])'.

    Parameters
    ----------
    agent
      A KappaAgent object.

    site
      Name of the site. Should be a name in agent.sites.

    state
      The site state, e.g. '.', '1', 2, etc.
    """

    def __init__(self, agent, site, state="."):
        if isinstance(agent, KappaAgent) and (site not in agent.sites):
            raise ValueError("%s has not site %s" % (agent, site))
        self.agent = agent
        self.site = site
        self.state = str(state)

    def _kappa(self):
        agent = self.agent
        if isinstance(agent, KappaAgent):
            agent = agent.name
        return "%s(%s[%s])" % (agent, self.site, self.state)


class KappaRule:
    """Class to represent a Kappa rule. Possibly very incomplete.

    Parameters
    ----------
    name
      Name of the rule.

    reactants
      List of KappaSiteState that are consumed by the rule.

    sense
      String arrow like '->' or '<->'.

    products
      List of KappaSiteState that are produced by the rule.

    rate
      Value indicating the frequency at which the rule happens. (or pair of
      values if sense is '<->'.
    """

    def __init__(self, name, reactants, sense, products, rate):
        self.name = name
        self.reactants = reactants
        self.products = products
        self.rate = rate
        self.sense = sense

    def _kappa(self):
        if isinstance(self.rate, (list, tuple)):
            rate = self.rate
        else:
            rate = str(self.rate)
        return "'%s' %s %s %s @ %s" % (
            self.name,
            ",".join(reactant._kappa() for reactant in self.reactants),
            self.sense,
            ",".join(product._kappa() for product in self.products),
            rate,
        )



""" dna_sequencing_viewer/__init__.py """

# __all__ = []

from .KappaClasses import KappaAgent, KappaSiteState, KappaRule
from .FormattedKappaError import FormattedKappaError
from .KappaModel import KappaModel
from .agents_graphs import (
    plot_snapshot_agent_nodes_graph,
    snapshot_agent_nodes_to_graph,
    plot_snapshot_agents,
)
from .plot_simulation_time_series import plot_simulation_time_series



import termcolor


def lines_subtext(text, start_line, end_line):
    """Return the text excerpt between the specified lines."""
    return "\n".join(text.split("\n")[start_line:end_line])


def split_text_in_three(text, start_line, start_chr, end_line, end_chr):
    """Split the text in three using 2 (line, column) control points."""
    lines = text.split("\n")
    before = "\n".join(lines[:start_line] + [lines[start_line][:start_chr]])
    if start_line == end_line:
        excerpt = lines[start_line][start_chr:end_chr]
    else:
        excerpt = "\n".join(
            [lines[start_line][start_chr:]]
            + ["\n".join(lines[start_line + 1 : end_line])]
            + [lines[end_line][:end_chr]]
        )
    after = "\n".join([lines[end_line][end_chr:]] + lines[end_line + 1 :])
    return before, excerpt, after


class FormattedKappaError(Exception):
    """Class to raise and pretty-print Kappa script syntax errors.

    When Kappa complains that the script doesn't make sense (through kappy),
    you can catch the error, pass it to FormattedKappaError.from_kappa_error()
    and reraise it. The resulting error will nicely print the errors in
    context.

    Examples
    --------

    >>> kappa_client = kappy.KappaStd()
    >>> kappa_client.add_model_string(model_string)
    >>> try:
    >>>     kappa_client.project_parse()
    >>> except kappy.KappaError as kappa_error:
    >>>     raise FormattedKappaError.from_kappa_error(kappa_error,
    >>>                                                model_string)
    """

    colors = {"error": "red", "warning": "orange"}

    @classmethod
    def _format_error_item(cls, error_item, model_string):
        color = cls.colors.get(error_item["severity"], "black")
        header = "[%s] %s" % (error_item["severity"], error_item["text"])
        colored_header = termcolor.colored(header, color, attrs=("bold",))
        rg = error_item["range"]
        start_line = rg["bline"] - 1
        start_chr = rg["bchr"]
        end_line = rg.get("eline", start_line) - 1
        end_chr = rg["echr"]
        before, excerpt, after = split_text_in_three(
            model_string, start_line, start_chr, end_line, end_chr
        )
        colored_excerpt = termcolor.colored(excerpt, color, attrs=("bold",))
        body_text = before + colored_excerpt + after
        body_text = lines_subtext(body_text, max(0, start_line - 2), end_line + 2)
        return colored_header + "\n\n" + body_text

    @classmethod
    def formatted_string(cls, error_items, model_string):
        """Create a FormattedKappaError from a Kappy project parsing error."""
        return "\n\n".join(
            cls._format_error_item(error_item, model_string)
            for error_item in error_items
        )

    @classmethod
    def from_kappa_error(cls, kappa_error, model_string):
        """Create a FormattedKappaError from a Kappy project parsing error."""
        return FormattedKappaError(
            cls.formatted_string(
                model_string=model_string, error_items=kappa_error.args[0]
            )
        )



import networkx as nx
import matplotlib.pyplot as plt
import numpy as np


def snapshot_agent_nodes_to_graph(nodes, with_ports=True):
    """Turn a simulation result into a networkx graph of a bio-complex.

    The "nodes" are from the representation of a snapshot agent.

    If with_port is true, the ports (= agents binding sites) will be
    represented as nodes in the graph.

    Examples:
    --------

    >>> sim_results = model.get_simulation_results()
    >>> n, nodes = sim_results['snapshots']['NAME']['snapshot_agents']
    >>> graph = snapshot_agent_nodes_to_graph(nodes, with_ports=True)
    """
    graph = nx.Graph()
    for i, node in enumerate(nodes):
        graph.add_node(i, node_type="agent", node_name=node["node_type"])
        for j, site in enumerate(node["node_sites"]):
            site_type, site_data = site["site_type"]
            graph.add_node((i, j), node_type=site_type, node_name=site["site_name"])
            graph.add_edge(i, (i, j), edge_type="port", weight=4)
            if site_type == "port":
                for link in site_data["port_links"]:
                    graph.add_edge((i, j), tuple(link), edge_type="link", weight=1)
    if not with_ports:
        for node_id, node in list(graph.nodes.items()):
            if node["node_type"] == "agent":
                node_name = node["node_name"]
                for neighbor in list(graph.neighbors(node_id)):
                    graph.remove_edge(node_id, neighbor)
                    graph = nx.contracted_nodes(graph, node_id, neighbor)
                graph.nodes[node_id]["node_name"] = node_name
    return graph


def plot_snapshot_agent_nodes_graph(
    graph, ax=None, positions=None, pos_seed=123, figsize=(4, 4), layout_method="FR"
):
    """Plot a graph of a complex agent from a snapshot.

    Parameters
    ----------
    graph
      The graph to be plotted.

    ax
      A matplotlib ax on which to draw the figure. If none is provided, a
      new figure and ax will be created and returned at the end.

    positions
      A dict {node: position}. A ``layout_method`` can be provided instead
      for automatic positioning.

    pos_seed
      A seed to freeze the positions when using semi-random positioning
      methods.

    figsize
      Dimensions of the figure in inches (if ax is not provided).

    layout_method
      Networkx layout method for the graph drawing.
      Either 'FR' (fruchterman_reingold), 'spectral' or 'spring'.
    """
    if layout_method == "FR":
        positions = nx.layout.fruchterman_reingold_layout(graph, seed=pos_seed)
    elif layout_method == "spectral":
        positions = nx.layout.spectral_layout(graph)
    elif layout_method == "spring":
        positions = nx.layout.spring_layout(graph, seed=pos_seed)
    else:
        raise ValueError("Unsupported layout_method %s" % layout_method)
    if ax is None:
        _, ax = plt.subplots(1, figsize=figsize)
    ax.axis("off")
    for edge in graph.edges():
        edge_type = graph.get_edge_data(*edge)["edge_type"]
        is_port = edge_type == "port"
        (x1, y1), (x2, y2) = positions[edge[0]], positions[edge[1]]
        ax.plot(
            [x1, x2],
            [y1, y2],
            color="#5d51da" if is_port else "black",
            lw=4 if is_port else 1,
        )
    for node, pos in positions.items():
        graph_node = graph.nodes[node]
        is_agent = graph_node["node_type"] == "agent"
        ax.text(
            pos[0],
            pos[1],
            graph_node["node_name"],
            ha="center",
            va="center",
            bbox=dict(
                boxstyle="round",
                facecolor="#e7e5f9" if is_agent else "white",
                edgecolor="grey",
            ),
            fontdict=dict(weight="bold" if is_agent else "normal"),
        )


def plot_snapshot_agents(
    agents,
    with_ports=True,
    columns="auto",
    ax_inches=3,
    layout_method="FR",
    freq_cutoff=0,
):
    """Plot graph representations of all complexes in a multi-ax figure.

    Parameters
    ----------
    agents
      Snapshot agents as provided in the result of ``get_simulation_results()``
      by ``simulation_results['snapshots']['SNAP_NAME']['snapshot_agents']``.

    with_ports
      If true, ports (= agents binding sites) will be represented on the
      figure.

    columns
      How many columns (= plots by lines) in the figure ? Keep to "auto"
      for a more-or-less squared figure.

    ax_inches
      Size of one figure in inches.

    layout_method
      Networkx layout method for the graph drawing.
      Either 'FR' (fruchterman_reingold), 'spectral' or 'spring'.

    freq_cutoff
      All complexes with a number of occurences lower than frew_cutoff in
      the snapshot will be ignored.

    Returns
    -------

    fig, axes
      Matplotlib "figure" and "axes" list of the multi-ax figure generated.
    """
    total_agents = sum(n for n, data in agents)
    # Normalize frequencies, filter, and sort agents.
    agents = sorted(
        [
            (100.0 * frequency / total_agents, nodes)
            for frequency, nodes in agents
            if (1.0 * frequency / total_agents) > freq_cutoff
        ],
        key=lambda a: -a[0],
    )
    if columns == "auto":
        columns = int(np.sqrt(len(agents)) + 1)
    rows = int(np.ceil(len(agents) / columns))
    figsize = (ax_inches * columns, ax_inches * rows)
    fig, axes = plt.subplots(rows, columns, figsize=figsize)
    for ax in axes.flatten():
        ax.axis("off")
    for ax, (frequency, nodes) in zip(axes.flatten(), agents):
        graph = snapshot_agent_nodes_to_graph(nodes, with_ports=with_ports)
        plot_snapshot_agent_nodes_graph(graph, ax=ax, layout_method=layout_method)
        ax.set_title("%.01f%%\n" % frequency)
    fig.tight_layout()
    fig.subplots_adjust(hspace=0.8, top=0.8)
    return fig, axes



import matplotlib.pyplot as plt


def plot_simulation_time_series(plots_data, ax=None):
    """Plot the time series data from a KappaModel simulation.

    Examples
    --------

    >>> model = KappaModel(...)
    >>> simulation_results = model.get_simulation_results()
    >>> ax = plot_simulation_time_series(simulation_results['plots'])
    >>> ax.figure.savefig('basic_example_time_series.png')
    """
    plots_data = dict(plots_data.items())
    times = plots_data.pop("[T]")
    if ax is None:
        _, ax = plt.subplots(1)
    for label, series in plots_data.items():
        ax.plot(times, series, label=label)
    ax.legend()
    return ax



from topkappy import (
    KappaModel,
    KappaAgent,
    KappaRule,
    KappaSiteState,
    plot_simulation_time_series,
    plot_snapshot_agents,
)

model = KappaModel(
    agents=[
        KappaAgent("A", ("a", "b")),
        KappaAgent("B", ("b", "c")),
        KappaAgent("C", ("c", "d")),
    ],
    rules=[
        KappaRule(
            "a.b",
            [KappaSiteState("A", "b", "."), KappaSiteState("B", "b", ".")],
            "->",
            [KappaSiteState("A", "b", "1"), KappaSiteState("B", "b", "1")],
            rate=0.5e-2,
        ),
        KappaRule(
            "b.c",
            [KappaSiteState("B", "c", "."), KappaSiteState("C", "c", ".")],
            "->",
            [KappaSiteState("B", "c", "1"), KappaSiteState("C", "c", "1")],
            rate=2.1e-2,
        ),
    ],
    initial_quantities={"A": 100, "B": 100, "C": 100,},
    duration=10,
    snapshot_times={"end": 10},
    plots=[KappaSiteState("B", "b", "."), KappaSiteState("C", "c", ".")],
)

simulation_results = model.get_simulation_results()

ax = plot_simulation_time_series(simulation_results["plots"])
ax.figure.savefig("basic_example_time_series.png")

agents_at_end = simulation_results["snapshots"]["end"]["snapshot_agents"]
fig, axes = plot_snapshot_agents(agents_at_end)
fig.savefig("basic_example_agents_graphs.png")



import ez_setup

ez_setup.use_setuptools()

from setuptools import setup, find_packages

exec(open("topkappy/version.py").read())  # loads __version__

setup(
    name="topkappy",
    version=__version__,
    author="Zulko",
    url="https://github.com/Edinburgh-Genome-Foundry/topkappy",
    description="Pythonic wrapper for the Kappa language (bio simulations)",
    long_description=open("pypi-readme.rst").read(),
    license="MIT",
    keywords="simulation biology modeling complex kappa binding",
    packages=find_packages(exclude="docs"),
    install_requires=["kappy", "networkx", "matplotlib", "numpy", "termcolor"],
)
